/* [xirc2ps_cs.c wk 14.04.97] (1.8 1997/06/26 14:24:26)
 * Xircom Creditcard Ethernet Adapter IIps driver
 *
 * This driver works for the CE2, CEM28 and CEM33 cards.
 * There is some support for CE3 type cards but they do not work yet.
 *
 * Written by Werner Koch (werner.koch@guug.de),
 * based on David Hints skeleton driver.
 *
 * Please report bugs to: "xircom-bugs@isil.d.shuttle.de"
 *
 * Xircom's usage of the CIS entry manfid.card is sometimes random, if your
 * card is not correctly deteced you can use an option to set the type of the
 * card:
 *	   module "xirc2ps_cs" opts "card_type=n"
 *   or
 *	   insmod xirc2ps_cs.o card_type=n
 *
 *   where n is:
 *		 1  for a CE2 card
 *		 2  for a CE2 card with modem (CEM28)
 *		 3  for a CE3 card
 *		 4  for a CE3 card with modem (CEM33)
 *
 *
 * Thanks to David Hints for the PCMCIA package, Donald Becker for some
 * advices, Xircom for giving me the specs, 4PC GmbH Duesseldorf for
 * providing a CEM33 and to all the people who tested the driver.
 *
 * For those, who are willing to help in the development of xircom drivers,
 * I have setup a mailing list:
 *     xircom-devel@isil.d.shuttle.de
 * To subscribe send a message containing the word "subscribe"
 * in the subject or somewhere at the beginning of a line to:
 *     xircom-devel-request@isil.d.shuttle.de
 *
 ************************************************************************
 * Copyright (c) 1997 by Werner Koch (dd9jn)
 *
 * This driver is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/*#define PCMCIA_DEBUG 4 */

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#ifdef MODULE
#define init_xirc2ps_cs init_module
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/in.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/bitops.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ciscode.h>

#ifndef MANFID_XIRCOM
  #define MANFID_XIRCOM 	   0x0105
  #define PRODID_XIRCOM_CE2	   0x010b
  #define PRODID_XIRCOM_CE3	   0x010a
  #define PRODID_XIRCOM_CEM28_1    0x110a     /* version "1.00" */
  #define PRODID_XIRCOM_CEM28	   0x110b
  #define PRODID_XIRCOM_CEM33	   0x110c
#endif

#include <pcmcia/ds.h>

/* Time in jiffies before concluding Tx hung */
#define TX_TIMEOUT	40

/****************
 * Some constants used to access the hardware
 */

/* Register offsets and value constans */
#define XIRCREG_CR  0	/* Command register (wr) */
enum xirc_cr {
    TransmitPacket = 0x01,
    SoftReset = 0x02,
    EnableIntr = 0x04,
    ForceIntr  = 0x08,
    ClearTxFIFO = 0x10,
    ClearRxOvrun = 0x20,
    RestartTx	 = 0x40
};
#define XIRCREG_ESR 0	/* Ethernet status register (rd) */
enum xirc_esr {
    FullPktRcvd = 0x01, /* full packet in receive buffer */
    PktRejected = 0x04, /* a packet has been rejected */
    TxPktPend = 0x08,	/* TX Packet Pending */
    IncorPolarity = 0x10,
    MediaSelect = 0x20	/* set if TP, clear if AUI */
};
#define XIRCREG_PR  1	/* Page Register select */
#define XIRCREG_EDP 4	/* Ethernet Data Port Register */
#define XIRCREG_ISR 6	/* Ethernet Interrupt Status Register */
enum xirc_isr {
    TxBufOvr = 0x01,	/* TX Buffer Overflow */
    PktTxed  = 0x02,	/* Packet Transmitted */
    MACIntr  = 0x04,	/* MAC Interrupt occured */
    TxResGrant = 0x08,	/* Tx Reservation Granted */
    RxFullPkt = 0x20,	/* Rx Full Packet */
    RxPktRej  = 0x40,	/* Rx Packet Rejected */
    ForcedIntr= 0x80	/* Forced Interrupt */
};
#define XIRCREG1_IMR0 12 /* Ethernet Interrupt Mask Register (on page 1)*/
#define XIRCREG1_IMR1 13
#define XIRCREG0_TSO  8  /* Transmit Space Open Register (on page 0)*/
#define XIRCREG0_TRS  10 /* Transmit reservation Size Register (page 0)*/
#define XIRCREG0_DO   12 /* Data Offset Register (page 0) (wr) */
#define XIRCREG0_RSR  12 /* Receive Status Register (page 0) (rd) */
enum xirc_rsr {
    PhyPkt = 0x01,	/* set:physical packet, clear: multicast packet */
    BrdcstPkt = 0x02,	/* set if it is a broadcast packet */
    PktTooLong = 0x04,	/* set if packet length > 1518 */
    AlignErr = 0x10,	/* incorrect CRC and last octet not complete */
    CRCErr = 0x20,	/* incorrect CRC and last octet is complete */
    PktRxOk = 0x80	/* received ok */
};
#define XIRCREG0_PTR 13 /* packets transmitted register (rd) */
#define XIRCREG0_RBC 14 /* receive byte count regsister (rd) */
#define XIRCREG1_ECR 14 /* ethernet configurationn register */
enum xirc_ecr {
    FullDuplex = 0x04,	/* enable full duplex mode */
    LongTPMode = 0x08,	/* adjust for longer lengths of TP cable */
    DisablePolCor = 0x10,/* disable auto polarity correction */
    DisableLinkPulse = 0x20, /* disable link pulse generation */
    DisableAutoTx = 0x40, /* disable auto-transmit */
};
#define XIRCREG2_RBS 8	/* receive buffer start register */
#define XIRCREG2_LED 10 /* LED Configuration register */
/* values for the leds:    Bits 2-0 for led 1
 *  0 disabled		   Bits 5-3 for led 2
 *  1 collision
 *  2 noncollision
 *  3 link_detected
 *  4 incor_polarity
 *  5 jabber
 *  6 auto_assertion
 *  7 rx_tx_activity
 */
#define XIRCREG4_GPR0 8 /* General Purpose Register 0 */
#define XIRCREG4_GPR1 9 /* General Purpose Register 1 */
#define XIRCREG4_BOV 10 /* Bonding Version Register */
#define XIRCREG4_LMA 12 /* Local Memory Address Register */
#define XIRCREG4_LMD 14 /* Local Memory Data Port */
/* MAC register can only by accessed with 8 bit operations */
#define XIRCREG40_CMD0 8    /* Command Register (wr) */
enum xirc_cmd { 	    /* Commands */
    Transmit = 0x01,
    EnableRecv = 0x04,
    DisableRecv = 0x08,
    Abort = 0x10,
    Online = 0x20,
    IntrAck = 0x40,
    Offline = 0x80
};
#define XIRCREG5_RHSA0	10  /* Rx Host Start Address */
#define XIRCREG40_RXST0 9   /* Receive Status Register */
#define XIRCREG40_TXST0 11  /* Transmit Status Register 0 */
#define XIRCREG40_TXST1 12  /* Transmit Status Register 10 */
#define XIRCREG40_RMASK0 13  /* Receive Mask Register */
#define XIRCREG40_TMASK0 14  /* Transmit Mask Register 0 */
#define XIRCREG40_TMASK1 15  /* Transmit Mask Register 0 */
#define XIRCREG42_SWC0	8   /* Software Configuration 0 */
#define XIRCREG42_SWC1	9   /* Software Configuration 1 */
#define XIRCREG42_BOC	10  /* Back-Off Configuration */
#define XIRCREG44_TDR0	8   /* Time Domain Reflectometry 0 */
#define XIRCREG44_TDR1	9   /* Time Domain Reflectometry 1 */
#define XIRCREG44_RXBC_LO 10 /* Rx Byte Count 0 (rd) */
#define XIRCREG44_RXBC_HI 11 /* Rx Byte Count 1 (rd) */
#define XIRCREG45_REV	 15 /* Revision Register (rd) */
#define XIRCREG50_IA	8   /* Individual Address (8-13) */


static char *if_names[] = { "Auto", "10BaseT", "10Base2", NULL };

/****************
 * All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
 * you do not define PCMCIA_DEBUG at all, all the debug code will be
 * left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
 * be present but disabled -- but it can then be enabled for specific
 * modules at load time with a 'pc_debug=#' option to insmod.
 */
#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#endif
static char *version =
"xirc2ps_cs.c 1.8 1997/06/26 14:24:26 (Werner Koch)";
	    /* !--- CVS revision */
#define KDBG_XIRC KERN_DEBUG   "xirc2ps_cs: "
#define KERR_XIRC KERN_ERR     "xirc2ps_cs: "
#define KWRN_XIRC KERN_WARNING "xirc2ps_cs: "
#define KNOT_XIRC KERN_NOTICE  "xirc2ps_cs: "
#define KINF_XIRC KERN_INFO    "xirc2ps_cs: "

/* card types */
#define XIR_UNKNOWN  0
#define XIR_CE2      1
#define XIR_CEM28    2
#define XIR_CE3      3
#define XIR_CEM33    4
/*====================================================================*/

/* Parameters that can be set with 'insmod' */
static int if_port = 1;
MODULE_PARM(if_port, "i");

/* Bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */
static u_long irq_mask = 0xdeb8;
MODULE_PARM(irq_mask, "i");

static int card_type = XIR_UNKNOWN; /* used to override the card_type */
MODULE_PARM(card_type, "i");

/*====================================================================*/

/* We do not process more than these number of bytes during one
 * interrupt. (Of course we receive complete packets, so this is not
 * an exact value).
 * Something between 2000..22000; first value gives best interrupt latency,
 * the second enables the usage of the complete on-chip buffer. We use the
 * high value as the initial value.
 */
static unsigned maxrx_bytes = 22000;


/*
 * The event() function is this driver's Card Services event handler.
 * It will be called by Card Services when an appropriate card status
 * event is received.  The config() and release() entry points are
 * used to configure or release a socket, in response to card insertion
 * and ejection events.  They are invoked from the event handler.
 */

static void xirc2ps_config(dev_link_t * link);
static void xirc2ps_release(u_long arg);
static int xirc2ps_event(event_t event, int priority,
			 event_callback_args_t * args);

/****************
 * The attach() and detach() entry points are used to create and destroy
 * "instances" of the driver, where each instance represents everything
 * needed to manage one actual PCMCIA card.
 */

static dev_link_t *xirc2ps_attach(void);
static void xirc2ps_detach(dev_link_t *);

/****************
 * You'll also need to prototype all the functions that will actually
 * be used to talk to your device.  See 'pcmem_cs' for a good example
 * of a fully self-sufficient driver; the other drivers rely more or
 * less on other parts of the kernel.
 */

void xirc2ps_interrupt IRQ(int irq, void *instance, struct pt_regs *regs);

/*
 * The dev_info variable is the "key" that is used to match up this
 * device driver with appropriate cards, through the card configuration
 * database.
 */

static dev_info_t dev_info = "xirc2ps_cs";

/****************
 * A linked list of "instances" of the device.  Each actual
 * PCMCIA card corresponds to one device instance, and is described
 * by one dev_link_t structure (defined in ds.h).
 *
 * You may not want to use a linked list for this -- for example, the
 * memory card driver uses an array of dev_link_t pointers, where minor
 * device numbers are used to derive the corresponding array index.
 */

static dev_link_t *dev_list = NULL;

/****************
 * A dev_link_t structure has fields for most things that are needed
 * to keep track of a socket, but there will usually be some device
 * specific information that also needs to be kept track of.  The
 * 'priv' pointer in a dev_link_t structure can be used to point to
 * a device-specific private data structure, like this.
 *
 * A driver needs to provide a dev_node_t structure for each device
 * on a card.  In some cases, there is only one device per card (for
 * example, ethernet cards, modems).  In other cases, there may be
 * many actual or logical devices (SCSI adapters, memory cards with
 * multiple partitions).  The dev_node_t structures need to be kept
 * in a linked list starting at the 'dev' field of a dev_link_t
 * structure.  We allocate them in the card's private data structure,
 * because they generally can't be allocated dynamically.
 */


typedef struct local_info_t {
    dev_node_t node;
    struct enet_statistics stats;
    int card_type;
    int media; /* detected media 1 := TP, 2 := ThinNet */
    int silicon; /* silicon revision. 0=old CE2, 1=Scipper, 4=Mohawk */
    int mohawk;  /* a CE3 type card */
    int multi;	 /* is a multi fucntion card (i.e with a modem) */
    int suspended;
    unsigned last_ptr_value; /* last packets transmitted value */
} local_info_t;




/****************
 * Some more prototypes
 */
static int do_start_xmit(struct sk_buff *skb, struct device *dev);
static struct enet_statistics *do_get_stats(struct device *dev);
static void set_addresses(struct device *dev);
#ifdef NEW_MULTICAST
static void set_multicast_list(struct device *dev);
#endif
static int do_init(struct device *dev);
static void set_card_type( local_info_t *local, int a, const char *s );
static int do_config(struct device *dev, struct ifmap *map);
static int do_open(struct device *dev);
static void do_reset(struct device *dev);
static void do_powerdown(struct device *dev);
static int do_close(struct device *dev);


/*=============== Helper functions =========================*/
static void
cs_error(int func, int ret)
{
    CardServices(ReportError, dev_info, (void *)func, (void *)ret);
}

static int
get_tuple_data(int fn, client_handle_t handle, tuple_t *tuple )
{
    int err;

    if( (err=CardServices(fn, handle, tuple)) )
	return err;
    return CardServices(GetTupleData, handle, tuple);
}

static int
get_tuple(int fn, client_handle_t handle, tuple_t *tuple, cisparse_t *parse)
{
    int err;

    if( (err=get_tuple_data(fn, handle, tuple)) )
	return err;
    return CardServices(ParseTuple, handle, tuple, parse);
}


#define first_tuple(a, b, c) get_tuple(GetFirstTuple, a, b, c)
#define next_tuple(a, b, c)  get_tuple(GetNextTuple, a, b, c)

#define SelectPage(pgnr)   outb((pgnr), ioaddr + XIRCREG_PR)
#define GetByte(reg)	   ((unsigned)inb(ioaddr + (reg)))
#define GetWord(reg)	   ((unsigned)inw(ioaddr + (reg)))
#define PutByte(reg,value) outb((value), ioaddr+(reg))
#define PutWord(reg,value) outw((value), ioaddr+(reg))

static void
busy_loop(u_long len)
{
    u_long timeout = jiffies + len;
    u_long flags;

    save_flags(flags);
    sti();
    while(timeout >= jiffies)
	;
    restore_flags(flags);
}


/*====== Functions used for debugging =================================*/
#if defined(PCMCIA_DEBUG) && 0
static void
PrintRegisters(struct device *dev)
{
    u_short ioaddr = dev->base_addr;

    if(pc_debug > 1) { /* FIXME: reading regs may change system status */
	int i, page;

	printk(KDBG_XIRC "Register  common: ");
	for(i = 0; i < 8; i++ )
	    printk(" %2.2x", GetByte(i) );
	printk("\n");
	for(page = 0; page <= 8; page++) {
	    printk(KDBG_XIRC "Register page %2x: ", page);
	    SelectPage(page);
	    for(i = 8; i < 16; i++)
		printk(" %2.2x", GetByte(i));
	    printk("\n");
	}
	for(page=0x40 ; page <= 0x5f; page++) {
	    if( page == 0x43 || (page >= 0x46 && page <= 0x4f)
		|| (page >= 0x51 && page <=0x5e) )
		continue;
	    printk(KDBG_XIRC "Register page %2x: ", page);
	    SelectPage(page);
	    for(i = 8; i < 16; i++)
		printk(" %2.2x", GetByte(i));
	    printk("\n");
	}
    }
}
#endif /* PCMCIA_DEBUG */

/*====== Real Functions  =================================*/




/****************
 * xirc2ps_attach() creates an "instance" of the driver, allocating
 * local data structures for one device.  The device is registered
 * with Card Services.
 *
 * The dev_link structure is initialized, but we don't actually
 * configure the card at this point -- we wait until we receive a
 * card insertion event.
 */

static dev_link_t *
xirc2ps_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    struct device *dev;
    local_info_t *local;
    int err;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "attach()\n");
  #endif

    /* Initialize the dev_link_t structure */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &xirc2ps_release;
    link->release.data = (u_long) link;


    /* General socket configuration */
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;

    /* Allocate space for a device structure */
    dev = kmalloc(sizeof(struct device), GFP_KERNEL);
    memset(dev, 0, sizeof(struct device));
    local = kmalloc(sizeof(local_info_t), GFP_KERNEL);
    memset(local, 0, sizeof(local_info_t));
    dev->priv = local;

    /* Fill in card specific entries */
    dev->hard_start_xmit = &do_start_xmit;
    dev->set_config = &do_config;
    dev->get_stats = &do_get_stats;
  #ifdef NEW_MULTICAST
    dev->set_multicast_list = &set_multicast_list;
  #else
    #warning "This driver does not support multicast with old kernels."
  #endif
    ether_setup(dev);
    dev->name = local->node.dev_name;
    dev->init = &do_init;
    dev->open = &do_open;
    dev->stop = &do_close;
    dev->tbusy = 1;
    link->priv = dev;

    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &xirc2ps_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    if( (err = CardServices(RegisterClient, &link->handle, &client_reg)) ) {
	cs_error(RegisterClient, err);
	xirc2ps_detach(link);
	return NULL;
    }

    return link;
} /* xirc2ps_attach */


/****************
 *  This deletes a driver "instance".  The device is de-registered
 *  with Card Services.  If it has been released, all local data
 *  structures are freed.  Otherwise, the structures will be freed
 *  when the device is released.
 */

static void
xirc2ps_detach(dev_link_t * link)
{
    dev_link_t **linkp;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "detach(0x%p)\n", link);
  #endif

    /* Locate device structure */
    for( linkp = &dev_list; *linkp; linkp = &(*linkp)->next )
	if( *linkp == link)
	    break;
    if( !*linkp ) {
      #ifdef PCMCIA_DEBUG
	printk(KDBG_XIRC "detach(0x%p): dev_link lost\n", link );
      #endif
	return;
    }

    /*
     * If the device is currently configured and active, we won't
     * actually delete it yet.	Instead, it is marked so that when
     * the release() function is called, that will trigger a proper
     * detach().
     */
    if(link->state & DEV_CONFIG) {
      #ifdef PCMCIA_DEBUG
	printk(KDBG_XIRC "detach postponed, '%s' "
			  "still locked\n", link->dev->dev_name);
      #endif
	link->state |= DEV_STALE_LINK;
	return;
    }

    /* Break the link with Card Services */
    if(link->handle)
	CardServices(DeregisterClient, link->handle);

    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if(link->priv) {
	struct device *dev = link->priv;
	if( dev->priv )
	    kfree_s(dev->priv, sizeof local_info_t );
	kfree_s(link->priv, sizeof(struct device));
    }
    kfree_s(link, sizeof(struct dev_link_t));

} /* xirc2ps_detach */


static void
set_card_type( local_info_t *local, int a, const char *s )
{
    local->mohawk = 0;
    switch( a ) {
      case PRODID_XIRCOM_CE2:
	printk(KNOT_XIRC "CE2 card %s\n",s );
	local->card_type = XIR_CE2;
	break;
      case PRODID_XIRCOM_CEM28:
	printk(KNOT_XIRC "CEM 28 card %s\n",s );
	local->card_type = XIR_CEM28;
	local->multi = 1;
	break;
      case PRODID_XIRCOM_CEM28_1:
	printk(KNOT_XIRC "CEM 28 card (v1) %s\n", s);
	local->card_type = XIR_CEM28;
	local->multi = 1;
	break;
      case PRODID_XIRCOM_CEM33:
	printk(KNOT_XIRC "CEM 33 card %s\n", s);
	local->card_type = XIR_CEM33;
	local->multi = 1;
	break;
      case PRODID_XIRCOM_CE3:
	printk(KNOT_XIRC "CE3 10/100 card %s\n", s);
	local->card_type = XIR_CE3;
	local->mohawk = 1;
	break;
      default:
	printk(KNOT_XIRC "warning: unknown card type\n");
	local->card_type = XIR_UNKNOWN;
    }
}

/****************
 * xirc2ps_config() is scheduled to run after a CARD_INSERTION event
 * is received, to configure the PCMCIA socket, and to make the
 * ethernet device available to the system.
 */
static void
xirc2ps_config(dev_link_t * link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
    struct device *dev;
    local_info_t *local;
    u_short ioaddr;
    int err, i;
    u_char buf[64];
    cistpl_lan_node_id_t *node_id = (cistpl_lan_node_id_t*)parse.funce.data;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;

    handle = link->handle;
    dev = link->priv;
    local = dev->priv;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "config(0x%p)\n", link);
  #endif

    /*
     * This reads the card's CONFIG tuple to find its configuration
     * registers.
     */
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = 64;
    tuple.TupleOffset = 0;

    /* Is this a xircom  card */
    tuple.DesiredTuple = CISTPL_MANFID;
    if( (err=first_tuple(handle, &tuple, &parse))) {
	printk(KNOT_XIRC "manfid not found in CIS\n");
	goto failure;
    }
    else {
      #ifdef PCMCIA_DEBUG
	if(pc_debug)
	    printk(KDBG_XIRC "manfid=%x prodid=%x\n",
		    (unsigned)parse.manfid.manf,
		    (unsigned)parse.manfid.card );
      #endif
	if( parse.manfid.manf != MANFID_XIRCOM ) {
	    printk(KNOT_XIRC "not a xircom card\n");
	    goto failure;
	}

	set_card_type(local,  parse.manfid.card, "detected" );
	/* Because the manfid.card is not always correct we have an option
	 * to override this */
	switch( card_type ) {
	  case XIR_UNKNOWN:
	    break; /* no override */
	  case XIR_CE2:
	    set_card_type(local, PRODID_XIRCOM_CE2  , "used" );
	    break;
	  case XIR_CEM28:
	    set_card_type(local, PRODID_XIRCOM_CEM28, "used" );
	    break;
	  case XIR_CEM33:
	    set_card_type(local, PRODID_XIRCOM_CEM33, "used" );
	    break;
	  case XIR_CE3:
	    set_card_type(local, PRODID_XIRCOM_CE3  , "used" );
	    break;
	  default:
	    printk(KNOT_XIRC "invalid card_type override requested\n");
	    goto failure;
	}
    }

    /* get configuration stuff */
    tuple.DesiredTuple = CISTPL_CONFIG;
    if( (err=first_tuple(handle, &tuple, &parse)))
	goto cis_error;
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* get the ethernet address from the CIS */
    tuple.DesiredTuple = CISTPL_FUNCE;
    for( err = first_tuple(handle, &tuple, &parse); !err;
			     err = next_tuple(handle, &tuple, &parse) ) {
	/* Once I saw two CISTPL_FUNCE_LAN_NODE_ID entries:
	 * the first one with a length of zero the second correct -
	 * so I skip all entries with length 0 */
	if( parse.funce.type == CISTPL_FUNCE_LAN_NODE_ID
	    && ((cistpl_lan_node_id_t *)parse.funce.data)->nb )
	    break;
    }
    if( err ) { /* not found: try to get the node-id from tuple 0x89 */
	tuple.DesiredTuple = 0x89;  /* data layout looks like tuple 0x22 */
	if( !(err = get_tuple_data(GetFirstTuple, handle, &tuple )) ) {
	    if( tuple.TupleDataLen == 8 && *buf == CISTPL_FUNCE_LAN_NODE_ID )
		memcpy( &parse, buf, 8 );
	    else
		err = -1;
	}
    }
    if( err ) {
	printk(KNOT_XIRC "node-id not found in CIS\n");
	goto failure;
    }
    node_id = (cistpl_lan_node_id_t *)parse.funce.data;
    if( node_id->nb != 6 ) {
	printk(KNOT_XIRC "malformed node-id in CIS\n");
	goto failure;
    }
    for( i=0; i < 6; i++ )
	dev->dev_addr[i] = node_id->id[i];

    /* Configure card */
    link->state |= DEV_CONFIG;

    link->io.IOAddrLines =10;
    link->io.NumPorts1 = local->multi? 18 : 16;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;

    link->irq.Attributes = IRQ_HANDLE_PRESENT;
    if( local->multi )
	link->irq.Attributes |= IRQ_TYPE_DYNAMIC_SHARING|IRQ_FIRST_SHARED;
    else
	link->irq.Attributes |= IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;
    link->irq.IRQInfo2 = irq_mask;
    link->irq.Handler = xirc2ps_interrupt;

    if( local->multi ) {
	link->io.NumPorts2 = 8;
	link->io.Attributes2 = IO_DATA_PATH_WIDTH_8;
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	for( err = first_tuple(handle, &tuple, &parse); !err;
			     err = next_tuple(handle, &tuple, &parse) ) {
	    if( cf->io.nwin > 0  &&  (cf->io.win[0].base & 0xf) == 8 ) {
		link->conf.ConfigIndex = cf->index ;
		link->io.BasePort2 = cf->io.win[0].base;
		for(ioaddr = 0x300; ioaddr < 0x400; ioaddr += 0x10) {
		    link->io.BasePort1 = ioaddr;
		    if( !(err=CardServices(RequestIO, link->handle, &link->io)) )
			goto port_found;
		}
		link->io.BasePort1 = 0;
		if( !(err=CardServices(RequestIO, link->handle, &link->io)) )
		    goto port_found;
	    }
	}
      port_found:
	if( err )
	    goto config_error;
    }
    else {
	for(ioaddr = 0x300; ioaddr < 0x400; ioaddr += 0x10) {
	    link->io.BasePort1 = ioaddr;
	    if( !(err=CardServices(RequestIO, link->handle, &link->io)) )
		break;
	}
	if( err ) { /* let the cs try */
	    link->io.BasePort1 = 0;
	    err=CardServices(RequestIO, link->handle, &link->io);
	}
	if( err ) {
	    cs_error(RequestIO, err);
	    goto config_error;
	}
    }

    /****************
     * Now allocate an interrupt line.	Note that this does not
     * actually assign a handler to the interrupt.
     */
    if( (err=CardServices(RequestIRQ, link->handle, &link->irq))) {
	cs_error(RequestIRQ, err);
	goto config_error;
    }

    /****************
     * This actually configures the PCMCIA socket -- setting up
     * the I/O windows and the interrupt mapping.
     */
    if( (err=CardServices(RequestConfiguration,
					link->handle, &link->conf)) ) {
	cs_error(RequestConfiguration, err);
	goto config_error;
    }

    /* The if_port symbol can be set when the module is loaded */
    if( if_port >= 1 && if_port <= 2 )
	dev->if_port = if_port;
    else
	printk(KERN_NOTICE "xirc2ps_cs: invalid if_port requested\n");

    /* we can now register the device with the net subsystem */
    dev->irq = link->irq.AssignedIRQ;
    dev->base_addr = link->io.BasePort1;
    dev->tbusy = 0;
    if( (err=register_netdev(dev)) ) {
	printk(KERN_NOTICE "xirc2ps_cs: register_netdev() failed\n");
	goto config_error;
    }

    link->state &= ~DEV_CONFIG_PENDING;
    link->dev = &local->node;

    /* give some infos about the hardware */
    printk(KERN_INFO "%s: Xircom, port %#3lx, irq %d, %s, hwaddr",
		      dev->name, (u_long)dev->base_addr,
		      (int)dev->irq, if_names[dev->if_port]);
    for(i = 0; i < 6; i++)
	printk("%c%02X", i?':':' ', dev->dev_addr[i]);
    printk("\n");
    if( local->card_type == XIR_CE3 )
	printk(KWRN_XIRC "Sorry, the CE3 type cards do not work yet\n");

    return;

  config_error:
    link->state &= ~DEV_CONFIG_PENDING;
    xirc2ps_release((u_long)link);
    return;

  cis_error:
    printk(KERN_NOTICE "xirc2ps_cs: unable to parse CIS\n");
  failure:
    link->state &= ~DEV_CONFIG_PENDING;
} /* xirc2ps_config */


/****************
 * After a card is removed, xirc2ps_release() will unregister the net
 * device, and release the PCMCIA configuration.  If the device is
 * still open, this will be postponed until it is closed.
 */
static void
xirc2ps_release( u_long arg)
{
    dev_link_t *link = (dev_link_t *) arg;
    struct device *dev = link->priv;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "release(0x%p)\n", link);
  #endif

    /*
     * If the device is currently in use, we won't release until it
     * is actually closed.
     */
    if(link->open) {
      #ifdef PCMCIA_DEBUG
	if(pc_debug)
	    printk(KDBG_XIRC "release postponed, '%s' "
			      "still open\n", link->dev->dev_name);
      #endif
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    if( link->dev ) {
	if( dev->irq )
	   irq2dev_map[dev->irq] = NULL;
	unregister_netdev(dev);
    }
    link->dev = NULL;
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;

    if(link->state & DEV_STALE_LINK)
	xirc2ps_detach(link);

} /* xirc2ps_release */

/*====================================================================*/

/****************
 * The card status event handler.  Mostly, this schedules other
 * stuff to run after an event is received.  A CARD_REMOVAL event
 * also sets some flags to discourage the net drivers from trying
 * to talk to the card any more.
 *
 * When a CARD_REMOVAL event is received, we immediately set a flag
 * to block future accesses to this device.  All the functions that
 * actually access the device should check this flag to make sure
 * the card is still present.
 */

static int
xirc2ps_event(event_t event, int priority,
	      event_callback_args_t * args)
{
    dev_link_t *link = args->client_data;
    struct device *dev = link->priv;
    local_info_t *lp = dev? dev->priv : NULL;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "event(%d)\n", (int)event );
  #endif

    switch (event) {
      case CS_EVENT_REGISTRATION_COMPLETE:
	 #ifdef PCMCIA_DEBUG
	  if(pc_debug)
	      printk(KDBG_XIRC "registration complete\n");
	 #endif
	  break;
      case CS_EVENT_CARD_REMOVAL:
	  link->state &= ~DEV_PRESENT;
	  if(link->state & DEV_CONFIG) {
	      dev->tbusy = 1; dev->start = 0;
	      link->release.expires = RUN_AT(HZ / 20);
	      add_timer(&link->release);
	  }
	  break;
      case CS_EVENT_CARD_INSERTION:
	  link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	  xirc2ps_config(link);
	  break;
      case CS_EVENT_PM_SUSPEND:
	  link->state |= DEV_SUSPEND;
	  /* Fall through... */
      case CS_EVENT_RESET_PHYSICAL:
	  if(link->state & DEV_CONFIG) {
	      if(link->open) {
		  dev->tbusy = 1; dev->start = 0;
		  lp->suspended=1;
		  do_powerdown(dev);
	      }
	      CardServices(ReleaseConfiguration, link->handle);
	  }
	  break;
      case CS_EVENT_PM_RESUME:
	  link->state &= ~DEV_SUSPEND;
	  /* Fall through... */
      case CS_EVENT_CARD_RESET:
	  if(link->state & DEV_CONFIG) {
	     CardServices(RequestConfiguration, link->handle, &link->conf);
	     if( link->open) {
		 do_reset(dev);
		 lp->suspended=0;
		 dev->tbusy = 0; dev->start = 1;
	     }
	  }
	  break;
    }
    return 0;
} /* xirc2ps_event */

/*====================================================================*/

/****************
 * This is the Interrupt service route.
 *
 * Note: We cannot use instance to initialize dev, because
 *	 RequestIRQ() does not set it (to support linux versions below 1.3.70).
 *	 So we must use the irq2dev_map[].
 */
void
xirc2ps_interrupt IRQ(int irq, void *instance, struct pt_regs *regs)
{
    struct device *dev = irq2dev_map[irq];
    local_info_t *lp;
    u_short ioaddr;
    u_char saved_page;
    unsigned bytes_rcvd;
    unsigned int_status, eth_status, rx_status, tx_status;
    unsigned rsr, pktlen;
    ulong start_ticks = jiffies; /* fixme: jiffies rollover every 497 days
				  * is this something to worry about?
				  * -- on a laptop?
				  */


    if( !dev ) {
	printk(KWRN_XIRC "interrupt %d for unknown device\n", irq);
	return;
    }
    if( dev->interrupt ) {
	printk(KERR_XIRC "re-entering isr on irq %d (dev=%p)\n", irq, dev);
	return;
    }
    dev->interrupt = 1;
    lp = dev->priv;
    ioaddr = dev->base_addr;
    if( lp->mohawk ) { /* must disable the interrupt */
	PutByte(XIRCREG_CR, 0 );
    }

  #ifdef PCMCIA_DEBUG
    if(pc_debug > 6)
	printk(KDBG_XIRC "interrupt %d at %#x.\n", irq, ioaddr);
  #endif


    saved_page = GetByte(XIRCREG_PR);
    /* Read the ISR to see whats the cause for the interrupt.
     * This also clears the interrupt flags on CE2 cards
     */
    int_status = GetByte(XIRCREG_ISR);
    if( int_status == 0xff ) { /* card may be ejected */
      #ifdef PCMCIA_DEBUG
	if( pc_debug )
	    printk(KDBG_XIRC "interrupt %d for dead card\n", irq );
      #endif
	goto leave;
    }
    int_status &= 0xf7; /* clear the reserved bit */
    eth_status = GetByte(XIRCREG_ESR);
    eth_status &= 0x7d; /* clear reserved bits */

    SelectPage(0x40);
    rx_status  = GetByte(XIRCREG40_RXST0);
    PutByte(XIRCREG40_RXST0, (~rx_status & 0xff) );
    tx_status = GetByte(XIRCREG40_TXST0);
    tx_status |= GetByte(XIRCREG40_TXST1) << 8;
    PutByte(XIRCREG40_TXST0, 0 );
    PutByte(XIRCREG40_TXST1, 0 );

  #ifdef PCMCIA_DEBUG
    if( pc_debug > 3 )
	printk(KDBG_XIRC "ISR=%#2.2x ESR=%#2.2x RSR=%#2.2x TSR=%#4.4x\n",
			  int_status, eth_status, rx_status, tx_status );
  #endif

    /***** receive section ******/
    bytes_rcvd = 0;
    SelectPage(0);
    while( eth_status & FullPktRcvd ) {
	rsr = GetByte(XIRCREG0_RSR);
	if( bytes_rcvd > maxrx_bytes && (rsr & PktRxOk) ) {
	    /* too many bytes received during this int, drop the rest of the
	     * packets */
	    lp->stats.rx_dropped++;
	    PutWord(XIRCREG0_DO, 0x8000 ); /* issue cmd: skip_rx_packet */
	}
	else if( rsr & PktRxOk ) {
	    struct sk_buff *skb;

	    pktlen = GetWord(XIRCREG0_RBC);
	    bytes_rcvd += pktlen;

	  #ifdef PCMCIA_DEBUG
	    if( pc_debug > 5 )
		printk(KDBG_XIRC "rsr=%#02x packet_length=%u\n", rsr, pktlen );
	  #endif

	    skb = ALLOC_SKB(pktlen+1); /* 1 extra byte, so we can use insw */
	    if( !skb ) {
	      #ifdef PCMCIA_DEBUG
		if( pc_debug )
		    printk(KNOT_XIRC "low memory, packet dropped (size=%u)\n",
								     pktlen );
	      #endif
		lp->stats.rx_dropped++;
	    }
	    else { /* okay get the packet */
		if( lp->silicon == 0  ) { /* work around a hardware bug */
		    unsigned rhsa; /* receive start address */

		    SelectPage(5);
		    rhsa = GetWord(XIRCREG5_RHSA0);
		    SelectPage(0);
		    rhsa += 3; /* skip control infos */
		    if( rhsa >= 0x8000 )
			rhsa = 0;
		    if( rhsa + pktlen > 0x8000 ) {
			unsigned i;
			#define BLOCK_INPUT(buf, len) \
			    for(i=0; i < len ; i--, rhsa++ ) { \
				buf[i] = GetByte(XIRCREG_EDP); \
				if( rhsa == 0x8000 ) {	       \
				    rhsa = 0;		       \
				    i--;		       \
				}			       \
			    }
			GET_PACKET(dev, skb, pktlen );
			#undef BLOCK_INPUT
		    }
		    else {
		       #define BLOCK_INPUT(buf, len) \
				insw(ioaddr+XIRCREG_EDP, buf, (len+1)>>1 )
			 GET_PACKET(dev, skb, pktlen );
		       #undef BLOCK_INPUT
		    }
		}
		else {
		   #define BLOCK_INPUT(buf, len) \
			    insw(ioaddr+XIRCREG_EDP, buf, (len+1)>>1 )
		     GET_PACKET(dev, skb, pktlen );
		   #undef BLOCK_INPUT
		}
		skb->dev = dev;
		netif_rx(skb);
		lp->stats.rx_packets++;
		if( !(rsr & PhyPkt) )
		    lp->stats.multicast++;
	    }
	    PutWord(XIRCREG0_DO, 0x8000 ); /* issue cmd: skip_rx_packet */
	}
	else {
	  #ifdef PCMCIA_DEBUG
	    if( pc_debug > 5 )
		printk("rsr=%#02x\n", rsr );
	  #endif
	}
	if( rsr & PktTooLong ) lp->stats.rx_frame_errors++;
	if( rsr & CRCErr )     lp->stats.rx_crc_errors++;
	if( rsr & AlignErr )   lp->stats.rx_fifo_errors++; /* okay ? */
	/* get the new ethernet status */
	eth_status = GetByte(XIRCREG_ESR);
	eth_status &= 0x7d; /* clear reserved bits */
    }
    if( rx_status & 0x10 ) { /* Receive overrun */
	lp->stats.rx_over_errors++;
	PutByte(XIRCREG_CR, ClearRxOvrun);
      #ifdef PCMCIA_DEBUG
	if( pc_debug > 3 )
	    printk(KDBG_XIRC "receive overrun cleared\n" );
      #endif
    }

    /***** transmit section ******/
    if( int_status & PktTxed ) {
	unsigned n, nn;

	n = lp->last_ptr_value;
	nn = GetByte(XIRCREG0_PTR);
	lp->last_ptr_value = nn;
	if( nn < n ) /* rollover */
	    lp->stats.tx_packets += 256 - n;
	else if( n == nn ) { /* happens sometimes - don't know why */
	  #ifdef PCMCIA_DEBUG
	    if( pc_debug )
		printk(KDBG_XIRC "PTR not changed?\n" );
	  #endif
	}
	else
	    lp->stats.tx_packets += lp->last_ptr_value - n;
	dev->tbusy = 0;
	mark_bh(NET_BH);  /* Inform upper layers. */
    }
    if( tx_status & 0x0002 ) {	/* Execessive collissions */
      #ifdef PCMCIA_DEBUG
	if( pc_debug )
	    printk(KDBG_XIRC "tx restarted due to execssive collissions\n" );
      #endif
	PutByte(XIRCREG_CR, RestartTx );  /* restart transmitter process */
    }
    if( tx_status & 0x0040 )
	lp->stats.tx_aborted_errors++;

    /* recalculate our work chunk so that we limit the duration of this
     * ISR to about 1/10 of a second.
     * Calculate only if we received a reasonable amount of bytes.
     */
    if( bytes_rcvd > 1000 ) {
	u_long duration = jiffies - start_ticks;

	if( duration >= HZ/10 ) { /* if more than about 1/10 second */
	    maxrx_bytes = (bytes_rcvd * (HZ/10)) / duration;
	    if( maxrx_bytes < 2000 )
		maxrx_bytes = 2000;
	    else if( maxrx_bytes > 22000 )
		maxrx_bytes = 22000;
	  #ifdef PCMCIA_DEBUG
	    if( pc_debug > 1)
		printk(KDBG_XIRC "set maxrx=%u (rcvd=%u ticks=%lu)\n",
				  maxrx_bytes, bytes_rcvd, duration  );
	  #endif
	}
	else if( !duration && maxrx_bytes < 22000 ) { /* now much faster*/
	    maxrx_bytes += 2000;
	    if( maxrx_bytes > 22000 )
		maxrx_bytes = 22000;
	  #ifdef PCMCIA_DEBUG
	    if( pc_debug > 1 )
		printk(KDBG_XIRC "set maxrx=%u\n", maxrx_bytes );
	  #endif
	}
    }

  leave:
    SelectPage(saved_page);
    dev->interrupt = 0;
    PutByte(XIRCREG_CR, EnableIntr );  /* re-enable interrupts */
    /* Instead of dropping packets during a receive, we could
     * force an interrupt with this command:
     *	  PutByte(XIRCREG_CR, EnableIntr|ForceIntr );
     * I have no dox about this command aside from it's name ---
     * The interrupt is kicked, but we must do something more do
     * clear it?
     * FIXME: the Mohawk dox says the Force Interrupt can be used ...
     */
} /* xirc2ps_interrupt */

/*====================================================================*/

static int
do_start_xmit(struct sk_buff *skb, struct device *dev)
{
    local_info_t *lp = dev->priv;
    u_short ioaddr = dev->base_addr;
    int okay;
    unsigned freespace;
    unsigned pktlen = skb? skb->len : 0;

  #ifdef PCMCIA_DEBUG
    if(pc_debug>1)
	printk(KDBG_XIRC "do_start_xmit(skb=%p, dev=%p) len=%u\n",
						    skb, dev, pktlen );
  #endif

    /* Transmitter timeout, serious problems */
    if( dev->tbusy ) {
	int tickssofar = jiffies - dev->trans_start;

	if( lp->suspended ) {
	    dev_kfree_skb (skb, FREE_WRITE);
	    dev->trans_start = jiffies;
	    lp->stats.tx_dropped++;
	    return 0;
	}
	if( tickssofar < TX_TIMEOUT )
	    return 1;

	printk(KNOT_XIRC "transmit timed out\n" );
	lp->stats.tx_errors++;
	/* reset the card */
	do_reset(dev);
	dev->trans_start = jiffies;
	dev->tbusy = 0;
    }

    if( !skb ) {
	dev_tint(dev);
	return 0;
    }

    if( skb->len <= 0 ) {
	dev_kfree_skb (skb, FREE_WRITE);
      #ifdef PCMCIA_DEBUG
	if( pc_debug )
	    printk(KDBG_XIRC "zero length packet dropped by tx\n");
      #endif
	return 0;
    }

    if( set_bit(0, (void*)&dev->tbusy ) ) {
	printk(KWRN_XIRC "transmitter access conflict\n");
	dev_kfree_skb (skb, FREE_WRITE);
	return 0;
    }

    /* adjust the packet length to min. required
     * and hope that the buffer is large enough
     * to provide some random data.
     */
    if( pktlen < ETH_ZLEN )
	pktlen = ETH_ZLEN;

    SelectPage(0);
    PutWord(XIRCREG0_TRS, (u_short)pktlen+2 );
    freespace = GetWord(XIRCREG0_TSO);
    okay = freespace & 0x8000;
    freespace &= 0x7fff;
    /* TRS doesn't work -(eliminated with sil-rev 1) */
    okay = pktlen +2 < freespace;
  #ifdef PCMCIA_DEBUG
    if(pc_debug > 2 )
	printk(KDBG_XIRC "avail. tx space=%u%s\n", freespace,
				okay? " (okay)":" (not enough)" );
  #endif
    if( !okay ) { /* not enough space */
	dev->tbusy = 1;
	return 1;  /* upper layer may decide to requeue this packet */
    }
    /* send the packet (fixme: could use 32 bit acces for CE3) */
    PutWord(XIRCREG_EDP, (ushort)pktlen );
    outsw(ioaddr+XIRCREG_EDP, skb->data, pktlen>>1 );
    if( pktlen & 1 )
	PutByte(XIRCREG_EDP, skb->data[pktlen-1] );
    if( lp->mohawk ) { /* must issue transmit command */
	PutByte(XIRCREG_CR, TransmitPacket );
    }

    dev_kfree_skb (skb, FREE_WRITE);
    dev->trans_start = jiffies;
    dev->tbusy = 0;
    return 0;
}

static struct enet_statistics *
do_get_stats(struct device *dev)
{
    local_info_t *lp = dev->priv;

    /*	lp->stats.rx_missed_errors = GetByte(?) */
    return &lp->stats;
}



/****************
 * Set all addresses: This first one is the individual address,
 * the next 9 addresses are taken from the multicast list and
 * the rest is filled with the individual address.
 */
static void
set_addresses(struct device *dev)
{
    u_short ioaddr = dev->base_addr;
    local_info_t *lp = dev->priv;
    struct dev_mc_list *dmi = dev->mc_list;
    char *addr;
    int i,j,k,n;


    SelectPage(k=0x50);
    for(i=0,j=8,n=0; ; i++, j++) {
	if( i > 5 ) {
	    if( ++n > 9 )
		break;
	    i = 0;
	}
	if( j > 15 ) {
	    j = 8;
	    k++;
	    SelectPage(k);
	}

	if( n && n <= dev->mc_count && dmi ) {
	    addr = dmi->dmi_addr;
	    dmi = dmi->next;
	}
	else
	    addr = dev->dev_addr;

	if( lp->mohawk )
	    PutByte( j, addr[5-i] );
	else
	    PutByte( j, addr[i] );
    }
    SelectPage(0);
}


/****************
 * Set or clear the multicast filter for this adaptor.
 * We can filter up to 9 addresses, if more are requested we set
 * multicast promiscuous mode.
 */

#ifdef NEW_MULTICAST
static void set_multicast_list(struct device *dev)
{
    u_short ioaddr = dev->base_addr;

    SelectPage(0x42);
    if( dev->flags & IFF_PROMISC ) { /* snoop */
	PutByte(XIRCREG42_SWC1, 0x06); /* set MPE and PME */
    }
    else if( dev->mc_count > 9 || (dev->flags & IFF_ALLMULTI) ) {
	PutByte(XIRCREG42_SWC1, 0x06); /* set MPE */
    }
    else if( dev->mc_count ) { /* the chip can filter 9 addresses perfectly */
	PutByte(XIRCREG42_SWC1, 0x00);
	SelectPage(0x40);
	PutByte(XIRCREG40_CMD0, Offline );
	set_addresses(dev);
	SelectPage(0x40);
	PutByte(XIRCREG40_CMD0, EnableRecv | Online );
    }
    else { /* standard usage */
	PutByte(XIRCREG42_SWC1, 0x00);
    }
    SelectPage(0);
}
#endif /* NEW_MULTICAST */


/****************
 * We never need to do anything when a IIps device is "initialized"
 * by the net software, because we only register already-found cards.
 */
static int
do_init(struct device *dev)
{
  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "do_init(%p)\n", dev );
  #endif
    return 0;
}


static int
do_config(struct device *dev, struct ifmap *map)
{
  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "do_config(%p)\n", dev );
  #endif
    return 0;
}



/****************
 * Open the driver
 */
static int
do_open(struct device *dev)
{
    local_info_t *lp = dev->priv;
    dev_link_t *link;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "do_open(%p)\n", dev );
  #endif

    /* Check that the PCMCIA card is still here. */
    for( link = dev_list; link; link = link->next )
	if( link->priv == dev )
	    break;
    /* Physical device present signature. */
    if( !DEV_OK(link) )
	return -ENODEV;


    /* okay */
    link->open++;
    MOD_INC_USE_COUNT;

    irq2dev_map[dev->irq] = dev;
    dev->interrupt = 0; dev->tbusy = 0; dev->start = 1;
    lp->suspended = 0;
    do_reset(dev);

    return 0;
}


static void
do_reset(struct device *dev)
{
    local_info_t *local = dev->priv;
    u_short ioaddr = dev->base_addr;
    unsigned value;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "do_reset(%p)\n", dev );
  #endif

    /* We do a hardreset at this point */
    SelectPage(4);
    PutByte(XIRCREG4_GPR1, 0);	     /* clear bit 0: power down */
    busy_loop(HZ/25);		     /* wait 40 msec */
    PutByte(XIRCREG4_GPR1, 1);	     /* set bit 0: power up */
    if( local->mohawk ) {
	/* set pin GP1 and GP2 to output  (0x0c)
	 * set GP1 to high to power up the 10Mhz chip (0x01)
	 * set GP2 to low to power up the ML6692      (0x00)
	 */
	busy_loop(HZ/50);	     /* wait 20 msec */
	PutByte(XIRCREG4_GPR0, 0x0d);
    }
    busy_loop(HZ/2);		     /* the anlog circuits need about 500ms */
    /* issue a software reset */
    PutByte(XIRCREG_CR, SoftReset ); /* set */
    busy_loop(HZ/50);		     /* wait 20 msec */
    PutByte(XIRCREG_CR, 0 );	     /* clear */
    busy_loop(HZ/25);		     /* wait 40 msec */

    local->last_ptr_value = 0;

    /* Enable auto-media detect. This is done by setting bit 2 of the GPR0 */
    SelectPage(4);
    if( !local->mohawk ) {
	PutByte(XIRCREG4_GPR0, 4);
	busy_loop(HZ/25);	    /* wait 20 msec to let it complete */
    }
    value = GetByte(XIRCREG_ESR);    /* read the ESR */
    local->media = value & MediaSelect ? 1 : 2;
  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "ESR after softreset: %#02X\n", value );
  #endif
    local->silicon = local->mohawk ? (GetByte(XIRCREG4_BOV) & 0x70) >> 4
				   : (GetByte(XIRCREG4_BOV) & 0x30) >> 4;

    /* configure the LEDs */
    SelectPage(2);
    if( local->media == 1 )  /* TP: Link and Activity */
	PutByte(XIRCREG2_LED, 0x3b );
    else		     /* Coax: Collision and Activity */
	PutByte(XIRCREG2_LED, 0x3a );

    /* setup the ECR */
    SelectPage(1);
    PutByte(XIRCREG1_IMR1, 1); /* Set TxUnderrunDetect */
    value = GetByte(XIRCREG1_ECR);
    PutByte(XIRCREG1_ECR, value); /* we use the default */

    /* Setup the configuration registers */
    SelectPage(0x42);
    PutByte(XIRCREG42_SWC1, 0x00); /* should handle media selection here */
    PutByte(XIRCREG42_SWC0, 0x20); /* disable source insertion */

    if( local->silicon != 1 ) {
	/* set the local memory dividing line.
	 * The comments in the sample code say that this is only
	 * settable with the scipper version 2 which is revision 0.
	 * Always for CE3 cards
	 */
	SelectPage(2);
	PutWord(XIRCREG2_RBS, 0x2000 );
    }

    /* fixme: acording to the dox we should reset the rx Overrun condition
     * and the txtransmitfifo.
     * But the defaults should be sufficient
     */

    set_addresses(dev);

    /* Hardware workaround:
     * The receive byte pointer after reset is off by 1 so we need
     * to move the offset pointer back to 0.
     */
    SelectPage(0);
    PutWord(XIRCREG0_DO, 0x2000 ); /* change offset command, off=0 */


    /* enable receiver and put the mac online */
    SelectPage(0x40);
    PutByte(XIRCREG40_CMD0, EnableRecv | Online );

    /* setup MAC IMRs and clear status registers */
    SelectPage(0x40);		     /* Bit 7 ... bit 0 */
    PutByte(XIRCREG40_RMASK0, 0xff); /* ROK, RAB, rsv, RO, CRC, AE, PTL, MP */
    PutByte(XIRCREG40_TMASK0, 0xff); /* TOK, TAB, SQE, LL, TU, JAB, EXC, CRS */
    PutByte(XIRCREG40_TMASK1, 0xb0); /* rsv, rsv, PTD, EXT, rsv,rsv,rsv, rsv*/
    PutByte(XIRCREG40_RXST0,  0x00); /* ROK, RAB, REN, RO, CRC, AE, PTL, MP */
    PutByte(XIRCREG40_TXST0,  0x00); /* TOK, TAB, SQE, LL, TU, JAB, EXC, CRS */
    PutByte(XIRCREG40_TXST1,  0x00); /* TEN, rsv, PTD, EXT, retry_counter:4  */


    /* setup Ethernet IMR and enable interrupts */
    SelectPage(0x01);
    PutByte(XIRCREG1_IMR0, 0xff ); /* allow all ints */
    SelectPage(0);
    PutByte(XIRCREG_CR, EnableIntr );
    if( local->multi ) { /* do some magic */
	if( !(GetByte( 0x10 ) & 0x01 ) )
	    PutByte( 0x10, 0x11 ); /* unmask master-int bit */
    }

  #ifdef PCMCIA_DEBUG
    if(pc_debug ) {
      #if 0 /* should not read out the registers because the receiver is online */
	SelectPage(0x50);
	printk(KDBG_XIRC "Ethernet address: ");
	for(i = 0; i < 6; i++)
	    printk("%c%02X", i?':':' ',
		    local->mohawk ? (unsigned)GetByte(XIRCREG50_IA +[5-i])
				  : (unsigned)GetByte(XIRCREG50_IA +i) );
	printk("\n");
      #endif
	printk(KDBG_XIRC "media %d, silicon revision %d\n",
					    local->media, local->silicon);
	/* We should switch back to page 0 to avoid a bug in revision 0
	 * where regs with offset below 8 can't be read after an access
	 * to the MAC registers */
	SelectPage(0);
    }
    /*PrintRegisters(dev);*/
  #endif
}

static void
do_powerdown(struct device *dev)
{

    u_short ioaddr = dev->base_addr;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "do_powerdown(%p)\n", dev );
  #endif

    SelectPage(4);
    PutByte(XIRCREG4_GPR1, 0);	     /* clear bit 0: power down */
    SelectPage(0);
}


static int
do_close( struct device *dev)
{
    u_short ioaddr = dev->base_addr;
    dev_link_t *link;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "do_close(%p)\n", dev );
  #endif

    for(link = dev_list; link; link = link->next)
	if(link->priv == dev)
	    break;
    if( !link )
	return -ENODEV;

  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "shutting down\n");
  #endif
    dev->tbusy = 1;
    dev->start = 0;

    SelectPage(0);
    PutByte(XIRCREG_CR, 0 );  /* disable interrupts */
    SelectPage(0x01);
    PutByte(XIRCREG1_IMR0, 0x00 ); /* forbid all ints */
    SelectPage(4);
    PutByte(XIRCREG4_GPR1, 0);	/* clear bit 0: power down */
    SelectPage(0);

    link->open--; dev->start = 0;
    if (link->state & DEV_STALE_CONFIG) {
	link->release.expires = RUN_AT(HZ/20);
	link->state |= DEV_RELEASE_PENDING;
	add_timer(&link->release);
    }

    MOD_DEC_USE_COUNT;

    return 0;
}



int
init_xirc2ps_cs(void)
{
    servinfo_t serv;

    printk(KERN_INFO "%s\n", version);
    CardServices(GetCardServicesInfo, &serv);
    if( serv.Revision != CS_RELEASE_CODE ) {
	printk(KNOT_XIRC "Card Services release does not match!\n");
	return -1;
    }
    register_pcmcia_driver(&dev_info, &xirc2ps_attach, &xirc2ps_detach);
    return 0;
}

#ifdef MODULE
void
cleanup_module(void)
{
  #ifdef PCMCIA_DEBUG
    if(pc_debug)
	printk(KDBG_XIRC "unloading\n");
  #endif
    unregister_pcmcia_driver(&dev_info);
    while( dev_list ) {
	if( dev_list->state & DEV_CONFIG )
	    xirc2ps_release( (u_long)dev_list );
	xirc2ps_detach( dev_list );
    }
}
#endif	/* MODULE */
